unit DList;

interface

uses
  SysUtils;

type

  TDCompare = function( A , B : Pointer ) : Integer;

  PDListNode = ^TDListNode;

  TDListNode = record
    Pred  : PDListNode;
    Succ  : PDListNode;
  end;

  EDListException = class( Exception ) end;
  EDListUnderflow = class( EDlistException ) end;

  TDList = class
  private
    FHead       : TDListNode;
    FPHead      : PDlistNode;
    FCompare    : TDCompare;
    FDataSize   : Integer;
    FOwn        : Boolean;
    FCount      : Integer;
  public
    constructor Create( DataSize : Integer; ACompare : TDCompare; Own : Boolean );
    destructor Destroy; override;
    procedure Clear;
    function IsEmpty : Boolean;
    function GetNodeData( Node : PDListNode ) : Pointer;
    procedure SetNodeData( Node : PDListNode; Data : Pointer );

    function InsertFirst( Data : Pointer ) : PDListNode;
    function InsertNodeFirst( Node : PDListNode ) : PDListNode;
    function InsertLast( Data : Pointer ) : PDListNode;
    function InsertNodeLast( Node : PDListNode ) : PDListNode;
    function Insert( Data : Pointer ) : PDListNode;
    function InsertNode( Node : PDListNode ) : PDListNode;
    function RemoveFirst : PDListNode;
    function RemoveLast : PDListNode;
    function Remove( Data : Pointer ) : PDListNode;
    function Find( Data : Pointer ) : PDListNode;

    procedure FreeNode( Node : PDListNode );
    function AllocateNode( Data : Pointer ) : PDListNode;

    function Detach( Node : PDListNode ) : PDListNode;
    function AttachAfter( ExistingNode , NewNode : PDListNode ) : PDListNode;

    function First : PDListNode;
    function Next( Node : PDListNode ) : PDListNode;
    function IsLast( Node : PDListNode ) : Boolean;
    function Last : PDListNode;
    function Prior( Node : PDListNode ) : PDListNode;
    function IsFirst( Node : PDListNode ) : Boolean;

    property Count : Integer read FCount;
  end;

implementation


constructor TDList.Create( DataSize : Integer; ACompare : TDCompare; Own : Boolean );
begin
  inherited Create;
  FPHead := Addr( FHead );
  FHead.Pred := FPHead;
  FHead.Succ := FPHead;
  FCompare := ACompare;
  FDataSize := DataSize;
  FOwn := Own;
  FCount := 0;
end;

destructor TDList.Destroy;
begin
  if FOwn then
    Clear;
  inherited;
end;

function TDList.IsEmpty : Boolean;
begin
  Result := IsLast( FPHead );
end;

procedure TDList.Clear;
begin
  if FOwn then
    while not IsEmpty do
      FreeNode( Detach( FHead.Succ ) )
  else
    while not IsEmpty do
      Detach( FHead.Succ );
end;

function TDList.Detach( Node : PDListNode ) : PDListNode;
begin
  Node.Pred.Succ := Node.Succ;
  Node.Succ.Pred := Node.Pred;
  Result := Node;
  Dec( FCount );
end;

function TDList.AttachAfter( ExistingNode , NewNode : PDListNode ) : PDListNode;
begin
  NewNode.Succ := ExistingNode.Succ;
  NewNode.Pred := ExistingNode;
  ExistingNode.Succ := NewNode;
  NewNode.Succ.Pred := NewNode;
  Result := NewNode;
  Inc( FCount );
end;

function TDList.GetNodeData( Node : PDListNode ) : Pointer;
begin
  Result := PChar( Node ) + SizeOf( TDListNode );
end;

procedure TDList.SetNodeData( Node : PDListNode; Data : Pointer );
begin
  Move( Data^ , GetNodeData( Node )^ , FDataSize );
end;

function TDList.AllocateNode( Data : Pointer ) : PDListNode;
begin
  GetMem( Result , SizeOf( TDListNode ) + FDataSize );
  SetNodeData( Result , Data );
end;

function TDList.InsertFirst( Data : Pointer ) : PDListNode;
begin
  Result := InsertNodeFirst( AllocateNode( Data ) );
end;

function TDList.InsertNodeFirst( Node : PDListNode ) : PDListNode;
begin
  Result := AttachAfter( FPHead , Node );
end;

function TDList.InsertLast( Data : Pointer ) : PDListNode;
begin
  Result := InsertNodeLast( AllocateNode( Data ) );
end;

function TDList.InsertNodeLast( Node : PDListNode ) : PDListNode;
begin
  Result := AttachAfter( FHead.Pred , Node );
end;

function TDList.Insert( Data : Pointer ) : PDListNode;
begin
  Result := InsertNode( AllocateNode( Data ) );
end;

function TDList.InsertNode( Node : PDListNode ) : PDListNode;
var
  Target  : PDListNode;
  Data    : Pointer;
begin
  if Assigned( FCompare ) then
    begin
      Target := First;
      Data := GetNodeData( Node );
      while Target <> nil do
        begin
          if FCompare( Data , GetNodeData( Target ) ) <= 0 then
            begin
              Result := AttachAfter( Target.Pred , Node );
              Exit;
            end;
          Target := Next( Target );
        end;
    end;
  Result := InsertNodeLast( Node );
end;

function TDList.IsLast( Node : PDListNode ) : Boolean;
begin
  Result := Node.Succ = FPHead;
end;

function TDList.IsFirst( Node : PDListNode ) : Boolean;
begin
  Result := Node.Pred = FPHead;
end;

function TDList.First : PDListNode;
begin
  Result := Next( FPHead );
end;

function TDList.Last : PDListNode;
begin
  Result := Prior( FPHead );
end;

function TDList.Next( Node : PDListNode ) : PDListNode;
begin
  if IsLast( Node ) then
    Result := nil
  else
    Result := Node.Succ;
end;

function TDList.Prior( Node : PDListNode ) : PDListNode;
begin
  if IsFirst( Node ) then
    Result := nil
  else
    Result := Node.Pred;
end;

function TDList.RemoveFirst : PDListNode;
begin
  if IsEmpty then
    raise EDListUnderFlow.Create( 'Cannot remove item from empty list' );
  Result := Detach( FHead.Succ );
end;

function TDList.RemoveLast : PDListNode;
begin
  if IsEmpty then
    raise EDListUnderFlow.Create( 'Cannot remove item from empty list' );
  Result := Detach( FHead.Pred );
end;

function TDList.Find( Data : Pointer ) : PDListNode;
begin
  if not Assigned( FCompare ) then
    Result := nil
  else
    begin
      Result := First;
      while Result <> nil do
        begin
          if FCompare( Data , GetNodeData( Result ) ) = 0 then
            Break;
          Result := Next( Result );
        end;
    end;
end;

function TDList.Remove( Data : Pointer ) : PDListNode;
begin
  Result := Find( Data );
  if Result <> nil then
    Detach( Result );
end;

procedure TDList.FreeNode( Node : PDListNode );
begin
  if Node <> nil then
    FreeMem( Node );
end;

end.
