{
	THyperSpreadsheet - Requires THyperGrid version 1.60 and above.
	------------------

		Copyright Pablo Pissanetzky
		http://www.pablop.demon.co.uk
		pablop@bigfoot.com
}

unit HgSpreadsheet;

interface

uses
  Classes , HgGrid , Calculator , Grids , Windows;

type

  THyperSpreadsheet = class;

  PHgFunctionDef = ^ THgFunctionDef;
  THgFunctionDef = record
    Name      : string;
    ArgCount  : Integer;
    ArgTypes  : TValueTypes;
  end;

  THyperCalculator = class( TCalculator )
  private
    FSpreadsheet : THyperSpreadSheet;
  private
    function GetValueAt( ACol , ARow : Longint; var ValueType : TValueType ) : string;
    function IdToCoord( const Identifier : string; var Coord : TGridCoord ) : Boolean;
    function IdtoRect( const Identifier : string; var Rect : TRect ) : Boolean;
    function BuiltInFunctionDef( const Name : string ) : PHgFunctionDef;
  protected
    function ResolveIdentifier( const Identifier : string;
      var ValueType : TValueType; var Value : string ) : Boolean; override;
    function LookupFunction( const FunctionName : string;
      var MinArgCount , MaxArgCount : Integer; var ValueTypes : TValueTypes ) : Boolean; override;
    function SolveFunction( const FunctionName : string;
      Arguments : TStrings; var ReturnType : TValueType; var Return : string ) : Boolean; override;
  public
    constructor Create( Spreadsheet : THyperSpreadSheet );
  end;

  THyperSpreadsheet = class( THyperGrid )
  private
    FCalculator     : THyperCalculator;
    FAutoCalculate  : Boolean;
    FCalculating    : Boolean;
  protected
    function SolveFormula( const Formula : string ) : string; override;
    constructor Create( AOwner : TComponent ); override;
    destructor Destroy; override;
    procedure DoCellEdited( ACol , ARow : Longint ); override;
  published
    property AutoCalculate : Boolean read FAutoCalculate write FAutoCalculate;
  end;

  procedure Register;

implementation

uses
  SysUtils;

procedure Register;
begin
  RegisterComponents( 'Marley' , [ THyperSpreadsheet ] );
end;


//------------------------------------------------------------------------------

const
  BuiltInFunctions : array [ 1..5 ] of ThgFunctionDef =
  (
    ( Name  : ':'       ; ArgCount : 2;   ArgTypes : [ vtIdentifier ] ),
    ( Name  : 'AVERAGE' ; ArgCount : 1;   ArgTypes : [ vtIdentifier ] ),
    ( Name  : 'COUNT'   ; ArgCount : 1;   ArgTypes : [ vtIdentifier ] ),
    ( Name  : 'SUM'     ; ArgCount : 1;   ArgTypes : [ vtIdentifier ] ),
    ( Name  : 'IF'      ; ArgCount : 3;   ArgTypes : [ vtIdentifier , vtString , vtNumber , vtEmpty ] )
  );

//------------------------------------------------------------------------------

constructor THyperCalculator.Create( Spreadsheet : THyperSpreadSheet );
begin
  inherited Create;
  FSpreadsheet := Spreadsheet;
  CustomOperators := ':';
end;

function THyperCalculator.IdToCoord( const Identifier : string; var Coord : TGridCoord ) : Boolean;
var
  ColPart , RowPart , ID : string;
  I : Integer;
begin
  ColPart := '';
  RowPart := '';
  I := 1;
  ID := Uppercase( Identifier );
  while I <= Length( ID ) do
    begin
      if ID[ I ] in [ 'A'..'Z' ]  then
        ColPart := ColPart + ID[ I ]
      else
        Break;
      Inc( I );
    end;
  while I <= Length( ID ) do
    begin
      if ID[ I ] in [ '0'..'9' ] then
        RowPart := RowPart + ID[ I ]
      else
        Break;
      Inc( I );
    end;
  Result := ( Length( ColPart ) > 0 ) and ( Length( RowPart ) > 0 );
  if Result then
    begin
      Coord.X := 0;
      for I := 1 to Length( ColPart ) do
        Inc( Coord.X , ( Ord( ColPart[ I ] ) - Ord( 'A' ) ) + ( ( I - 1 ) * ( Ord( 'Z' ) - Ord( 'A' ) + 1 ) ) );
      Coord.Y := StrToInt( RowPart );
      Result := ( Coord.X >= 0 ) and ( Coord.X < FSpreadsheet.Columns.Count ) and
        ( Coord.Y >= 0 ) and ( Coord.Y < FSpreadsheet.RowCount );
    end;
end;

function THyperCalculator.IdtoRect( const Identifier : string; var Rect : TRect ) : Boolean;
var
  ID  : string;
  C   : TGridCoord;
  I : Integer;
begin
  ID := Identifier;
  Result := IdToCoord( ID , C );
  if Result then
    begin
      Rect.Left := C.X;
      Rect.Right := C.X;
      Rect.Top := C.Y;
      Rect.Bottom := C.Y;
      I := Pos( ' ' , ID );
      if I > 0 then
        begin
          Delete( ID , 1 , I );
          Result := IdToCoord( Trim( ID ) , C );
          if Result then
            begin
              if C.X <= Rect.Left then Rect.Left := C.X
              else Rect.Right := C.X;

              if C.Y <= Rect.Top then Rect.Top := C.Y
              else Rect.Bottom := C.Y;
            end;
        end;
    end;
end;

function THyperCalculator.GetValueAt( ACol , ARow : Longint; var ValueType : TValueType ) : string;
begin
  if not FSpreadsheet.Columns[ ACol ].Visible then
    Result := FSpreadsheet.Columns[ ACol ].Cols[ ARow ]
  else
    Result := FSpreadsheet.Cells[ FSpreadsheet.GridColumnForColumnObject( FSpreadsheet.Columns[ ACol ] ) , ARow ];

  Result := FSpreadsheet.SolveFormula( Result );
  if Length( Result ) = 0 then
    begin
      Result := '';
      ValueType := vtEmpty;
    end
  else
    try
      StrToFloat( Result );
      ValueType := vtNumber;
    except
      ValueType := vtString;
    end;
end;

function THyperCalculator.ResolveIdentifier( const Identifier : string;
  var ValueType : TValueType; var Value : string ) : Boolean;
var
  Coord : TGridCoord;
begin
  Result := IDToCoord( Identifier , Coord );
  if Result then
    Value := GetValueAt( Coord.X , Coord.Y , ValueType );
end;

function THyperCalculator.BuiltInFunctionDef( const Name : string ) : PHgFunctionDef;
var
  I : Integer;
begin
  Result := nil;
  for I := Low( BuiltInFunctions ) to High( BuiltInFunctions ) do
    begin
      if EqualStr( Name , BuiltInFunctions[ I ].Name ) then
        begin
          Result := @ BuiltInFunctions[ I ];
          Break;
        end;
    end;
end;


function THyperCalculator.LookupFunction( const FunctionName : string;
  var MinArgCount , MaxArgCount : Integer; var ValueTypes : TValueTypes ) : Boolean;
var
  FunctionDef : PHgFunctionDef;
begin
  Result := inherited LookupFunction( FunctionName , MinArgCount , MaxArgCount , ValueTypes );
  if not Result then
    begin
      FunctionDef := BuiltInFunctionDef( FunctionName );
      if FunctionDef = nil then
        Result := False
      else
        begin
          MinArgCount := FunctionDef.ArgCount;
          MaxArgCount := FunctionDef.ArgCount;
          ValueTypes := FunctionDef.ArgTypes;
          Result := True;
        end;
    end;
end;

function THyperCalculator.SolveFunction( const FunctionName : string;
  Arguments : TStrings; var ReturnType : TValueType; var Return : string ) : Boolean;
var
  Range     : TRect;
  C , R     : Longint;
  Number    : Double;
  ValueType : TValueType;
  Text      : string;

  function SumFunc : Double;
  var
    C , R : Integer;
  begin
    Result := 0;
    for C := Range.Left to Range.Right do
      for R := Range.Top to Range.Bottom do
        begin
          Text := GetValueAt( C , R , ValueType );
          if ValueType = vtNumber then
            Result := Result + StrToFloat( Text );
        end;
  end;

  function CountFunc : Double;
  begin
    Result := ( Range.Right - Range.Left + 1 ) * ( Range.Bottom - Range.Top + 1 );
  end;

  function AverageFunc : Double;
  begin
    Result := SumFunc / CountFunc;
  end;

var
  FunctionDef : PHgFunctionDef;
begin
  Result := inherited SolveFunction( FunctionName , Arguments , ReturnType , Return );
  if Result then
    Exit;

  case StringEval( FunctionName , [ ':' , 'SUM' , 'COUNT' , 'AVERAGE' ,'IF' ] ) of
    0:      {:}
      begin
        Result := True;
        ReturnType := vtIdentifier;
        Return := Arguments[ 0 ] + ' ' + Arguments[ 1 ];
      end;
    1:      {SUM}
      begin
        Result := IdToRect( Arguments[ 0 ] , Range );
        if Result then
          begin
            ReturnType := vtNumber;
            Return := FloatToStr( SumFunc );
          end;
      end;
    2:      {COUNT}
      begin
        Result := IdToRect( Arguments[ 0 ] , Range );
        if Result then
          begin
            ReturnType := vtNumber;
            Return := FloatToStr( CountFunc );
          end;
      end;
    3:      {AVERAGE}
      begin
        Result := IdToRect( Arguments[ 0 ] , Range );
        if Result then
          begin
            ReturnType := vtNumber;
            Return := FloatToStr( AverageFunc );
          end;
      end;
    4:      {IF}
      begin
        FunctionDef := BuiltInFunctionDef( FunctionName );
        if FunctionDef <> nil then
          begin
            ResolveArguments( Arguments , FunctionDef.ArgTypes );
            if ArgumentType( 0 , Arguments ) = vtString then
              begin
                Result := True;
                if Arguments[ 0 ] = 'True' then
                  begin
                    ReturnType := ArgumentType( 1 , Arguments );
                    Return := Arguments[ 1 ];
                  end
                else
                  begin
                    ReturnType := ArgumentType( 2 , Arguments );
                    Return := Arguments[ 2 ];
                  end;
              end;
          end;
      end;
  end;
end;

//------------------------------------------------------------------------------

constructor THyperSpreadSheet.Create( AOwner : TComponent );
begin
  inherited;
  FCalculator := THyperCalculator.Create( Self );
  FCalculating := False;
  FAutoCalculate := True;
end;

destructor THyperSpreadSheet.Destroy;
begin
  FCalculator.Free;
  inherited;
end;

function THyperSpreadsheet.SolveFormula( const Formula : string ) : string;
var
  Text : string;
  ReturnType : TValueType;
begin
  Text := Trim( Formula );
  if ( Length( Text ) > 0 ) and ( Text[ 1 ] = '=' ) then
    begin
      Delete( Text , 1 , 1 );
      try
        Result := FCalculator.Solve( Text , ReturnType );
      except
        Result := '#ERROR';
      end;
    end
  else
    Result := Formula;
end;

procedure THyperSpreadsheet.DoCellEdited( ACol , ARow : Longint );
begin
  if FAutoCalculate then
    Refresh;
  inherited;
end;

end.

